import logging
from unittest.mock import Mock

import pytest

from zino.trapd import TrapMessage
from zino.trapobservers.logged_traps import RestartTrapLogger


class TestRestartTrapLogger:
    @pytest.mark.parametrize("trap_name", ["coldStart", "warmStart"])
    @pytest.mark.asyncio
    async def test_when_handle_trap_is_called_it_should_log_trap_name(
        self, caplog, localhost_trap_originator, trap_name
    ):
        observer = RestartTrapLogger(state=Mock())
        trap = TrapMessage(agent=localhost_trap_originator, mib="SNMPv2-MIB", name=trap_name)
        with caplog.at_level(logging.INFO):
            await observer.handle_trap(trap=trap)
            assert f"localhost: {trap_name}" in caplog.text



import logging

from zino.tasks.errors import DeviceUnreachableError

_log = logging.getLogger(__name__)


async def run_all_tasks(device, state):
    try:
        await run_registered_tasks(device, state)
    except DeviceUnreachableError:
        _log.debug(f"Device {device.name} could not be reached. Any remaining tasks have been cancelled.")


async def run_registered_tasks(device, state):
    for task_class in get_registered_tasks():
        task = task_class(device, state)
        await task.run()


def get_registered_tasks():
    from zino.tasks.addrs import AddressMapTask
    from zino.tasks.bfdtask import BFDTask
    from zino.tasks.bgpstatemonitortask import BGPStateMonitorTask
    from zino.tasks.juniperalarmtask import JuniperAlarmTask
    from zino.tasks.linkstatetask import LinkStateTask
    from zino.tasks.reachabletask import ReachableTask
    from zino.tasks.vendor import VendorTask

    return [ReachableTask, VendorTask, AddressMapTask, LinkStateTask, BFDTask, BGPStateMonitorTask, JuniperAlarmTask]



"""Zino configuration models"""

from ipaddress import IPv4Address, IPv6Address
from os import R_OK, access
from os.path import isfile
from typing import Optional, Union

from pydantic import BaseModel, ConfigDict
from pydantic.functional_validators import AfterValidator
from typing_extensions import Annotated

DEFAULT_INTERVAL_MINUTES = 5
STATE_FILENAME = "zino-state.json"
EVENT_DUMP_DIR = "old-events"
POLLFILE = "polldevs.cf"

IPAddress = Union[IPv4Address, IPv6Address]


def validate_file_can_be_opened(filename: str) -> str:
    assert isfile(filename) and access(filename, R_OK), f"File {filename} doesn't exist or isn't readable"
    return filename


ExistingFileName = Annotated[str, AfterValidator(validate_file_can_be_opened)]


# config fields and default values from
# https://gitlab.sikt.no/verktoy/zino/blob/master/common/config.tcl#L18-44
class PollDevice(BaseModel):
    """Defines the attributes Zino needs/wants to be able to poll a device"""

    name: str
    address: IPAddress
    community: str = "public"
    dns: str = None
    interval: int = DEFAULT_INTERVAL_MINUTES
    ignorepat: Optional[str] = None
    watchpat: Optional[str] = None
    priority: int = 100
    timeout: int = 5
    retries: int = 3
    domain: str = None
    statistics: bool = True
    hcounters: bool = False
    do_bgp: bool = True
    port: int = 161


class Archiving(BaseModel):
    model_config = ConfigDict(extra="forbid")

    old_events_dir: str = EVENT_DUMP_DIR


class Authentication(BaseModel):
    model_config = ConfigDict(extra="forbid")

    file: ExistingFileName = "secrets"


class Persistence(BaseModel):
    model_config = ConfigDict(extra="forbid")

    file: str = STATE_FILENAME
    period: int = DEFAULT_INTERVAL_MINUTES


class Polling(BaseModel):
    model_config = ConfigDict(extra="forbid")

    file: ExistingFileName = POLLFILE
    period: int = 1


class Configuration(BaseModel):
    """Class for keeping track of the configuration set by zino.toml"""

    # throw ValidationError on extra keys
    model_config = ConfigDict(extra="forbid")

    archiving: Archiving = Archiving()
    authentication: Authentication = Authentication()
    persistence: Persistence = Persistence()
    polling: Polling = Polling()



"""Tests for the various helper/development binaries and scripts in the zino package"""

import os
import subprocess


def test_zino_help_screen_runs_without_error():
    assert subprocess.check_call(["zino", "--help"]) == 0


def test_getuptime_help_runs_without_error(polldevs_conf, monkeypatch):
    monkeypatch.chdir(os.path.dirname(polldevs_conf))
    assert subprocess.check_call(["python", "-m", "zino.getuptime", "-h"]) == 0


def test_polltest_help_runs_without_error(polldevs_conf, monkeypatch):
    monkeypatch.chdir(os.path.dirname(polldevs_conf))
    assert subprocess.check_call(["python", "-m", "zino.polltest", "-h"]) == 0



from unittest.mock import patch

import pytest

from zino.config.models import PollDevice
from zino.oid import OID
from zino.state import ZinoState
from zino.statemodels import InterfaceState, Port
from zino.tasks.linkstatetask import (
    BaseInterfaceRow,
    CollectedInterfaceDataIsNotSaneError,
    LinkStateTask,
    MissingInterfaceTableData,
)


class TestLinkStateTask:
    @pytest.mark.asyncio
    async def test_run_should_not_create_event_if_links_are_up(self, linkstatetask_with_links_up):
        task = linkstatetask_with_links_up
        assert (await task.run()) is None
        assert len(task.state.events) == 0

    @pytest.mark.asyncio
    async def test_run_should_create_event_if_at_least_one_link_is_down(self, linkstatetask_with_one_link_down):
        task = linkstatetask_with_one_link_down
        assert (await task.run()) is None
        assert len(task.state.events) == 1

    def test_when_patterns_are_empty_interface_should_not_be_ignored(self, task_with_dummy_device):
        data = BaseInterfaceRow(
            index=2, descr="GigabitEthernet1/2", alias="uplink", admin_status="up", oper_status="up", last_change=0
        )
        assert task_with_dummy_device._is_interface_watched(data)

    def test_when_interface_matches_watchpat_it_should_not_be_ignored(self, task_with_dummy_device):
        data = BaseInterfaceRow(
            index=2, descr="GigabitEthernet1/2", alias="uplink", admin_status="up", oper_status="up", last_change=0
        )
        task_with_dummy_device.device.watchpat = "Gigabit"
        assert task_with_dummy_device._is_interface_watched(data)

    def test_when_interface_doesnt_match_watchpat_it_should_be_ignored(self, task_with_dummy_device):
        data = BaseInterfaceRow(
            index=2, descr="GigabitEthernet1/2", alias="uplink", admin_status="up", oper_status="up", last_change=0
        )
        task_with_dummy_device.device.watchpat = "TenGiga"
        assert not task_with_dummy_device._is_interface_watched(data)

    def test_when_interface_matches_ignorepat_it_should_be_ignored(self, task_with_dummy_device):
        data = BaseInterfaceRow(
            index=2, descr="GigabitEthernet1/2", alias="uplink", admin_status="up", oper_status="up", last_change=0
        )
        task_with_dummy_device.device.ignorepat = ".*Ethernet"
        assert not task_with_dummy_device._is_interface_watched(data)

    def test_when_interface_state_is_missing_update_state_should_raise_exception(self, task_with_dummy_device):
        row = BaseInterfaceRow(index=42, descr="x", alias="x", admin_status="x", oper_status="x", last_change=0)
        port = Port(ifindex=42)
        empty_state_row = {}

        with pytest.raises(MissingInterfaceTableData):
            task_with_dummy_device._update_state(data=row, port=port, row=empty_state_row)

    def test_when_interface_data_is_empty_update_single_interface_should_raise_exception(self, task_with_dummy_device):
        with pytest.raises(CollectedInterfaceDataIsNotSaneError):
            task_with_dummy_device._update_single_interface({})

    def test_when_interface_data_is_empty_update_interfaces_should_keep_processing(self, task_with_dummy_device):
        assert (
            task_with_dummy_device._update_interfaces(
                {
                    OID(".1"): {},
                    OID(".2"): {},
                    OID(".3"): {},
                }
            )
            is None
        )


class TestBaseInterfaceRow:
    def test_when_index_is_missing_is_sane_should_return_false(self):
        row = BaseInterfaceRow(index=None, descr="x", alias="x", admin_status="x", oper_status="x", last_change=0)
        assert not row.is_sane()

    def test_when_descr_is_missing_is_sane_should_return_false(self):
        row = BaseInterfaceRow(index=42, descr=None, alias="x", admin_status="x", oper_status="x", last_change=0)
        assert not row.is_sane()

    def test_when_descr_and_index_are_present_is_sane_should_return_true(self):
        row = BaseInterfaceRow(index=42, descr="x", alias="x", admin_status="x", oper_status="x", last_change=0)
        assert row.is_sane()

    @pytest.mark.asyncio
    async def test_poll_single_interface_should_update_state(self, linkstatetask_with_one_link_down):
        target_index = 2
        await linkstatetask_with_one_link_down.poll_single_interface(target_index)
        device_state = linkstatetask_with_one_link_down.state.devices.get(linkstatetask_with_one_link_down.device.name)

        assert target_index in device_state.ports, f"no state for port {target_index} was stored"
        port = device_state.ports[target_index]
        assert port.state == InterfaceState.DOWN
        assert port.ifdescr == "2"
        assert port.ifalias == "from a famous"

    @pytest.mark.asyncio
    async def test_when_ifindex_is_1_poll_single_interface_should_not_crash(self, linkstatetask_with_one_link_down):
        """Regression test.  poll_single_interface constructs a GET-NEXT request by asking for the set of next values
        after ifindex-1 - but 0 indexes are illegal when encoding OIDs into tables.
        """
        target_index = 1
        assert await linkstatetask_with_one_link_down.poll_single_interface(target_index) is None

    @pytest.mark.asyncio
    async def test_when_timeout_occurs_poll_single_interface_should_not_crash(self, linkstatetask_with_one_link_down):
        with patch.object(linkstatetask_with_one_link_down.snmp, "getnext2") as mock_getnext2:
            mock_getnext2.side_effect = TimeoutError
            assert await linkstatetask_with_one_link_down.poll_single_interface(42) is None


@pytest.fixture
def linkstatetask_with_links_up(snmpsim, snmp_test_port):
    device = PollDevice(name="buick.lab.example.org", address="127.0.0.1", port=snmp_test_port, community="linksup")
    state = ZinoState()
    task = LinkStateTask(device, state)
    yield task


@pytest.fixture
def linkstatetask_with_one_link_down(snmpsim, snmp_test_port):
    device = PollDevice(name="buick.lab.example.org", address="127.0.0.1", port=snmp_test_port, community="linksdown")
    state = ZinoState()
    task = LinkStateTask(device, state)
    yield task


@pytest.fixture
def task_with_dummy_device():
    device = PollDevice(name="test", address="127.0.0.1")
    state = ZinoState()
    task = LinkStateTask(device, state)
    yield task



import asyncio
import logging
import operator
from datetime import datetime, timedelta
from typing import Sequence, Set, Tuple

from apscheduler.executors.asyncio import AsyncIOExecutor
from apscheduler.jobstores.base import JobLookupError
from apscheduler.schedulers.asyncio import AsyncIOScheduler

from zino import state
from zino.config.models import DEFAULT_INTERVAL_MINUTES, PollDevice
from zino.config.polldevs import InvalidConfiguration, read_polldevs
from zino.tasks import run_all_tasks

_log = logging.getLogger(__name__)
_scheduler = None


def get_scheduler() -> AsyncIOScheduler:
    """Returns the global scheduler.

    Creates the global scheduler on the first call.
    """
    global _scheduler
    if _scheduler is None:
        executors = {
            "default": AsyncIOExecutor(),
        }
        job_defaults = {
            "max_instances": 1,  # Never allow same job to run simultaneously
        }
        _scheduler = AsyncIOScheduler(
            event_loop=asyncio.get_event_loop(),
            executors=executors,
            job_defaults=job_defaults,
        )
    return _scheduler


def load_polldevs(polldevs_conf: str) -> Tuple[Set, Set]:
    """Loads pollfile into process state.

    :returns: A tuple of (new_devices, deleted_devices)
    """
    try:
        devices = {d.name: d for d in read_polldevs(polldevs_conf)}
    except InvalidConfiguration as error:
        _log.error(error)
        return set(), set()

    new_devices = set(devices) - set(state.polldevs)
    deleted_devices = set(state.polldevs) - set(devices)
    if new_devices:
        _log.info("loaded new devices: %r", new_devices)
        init_state_for_devices((devices[d] for d in new_devices))
    if deleted_devices:
        _log.info("deleted devices: %r", deleted_devices)

    state.polldevs.update(devices)
    for device in deleted_devices:
        del state.polldevs[device]

    return new_devices, deleted_devices


def init_state_for_devices(devices: Sequence[PollDevice]):
    """Initializes empty state structures for new devices, if none already exist"""
    for device in devices:
        state.state.addresses[device.address] = device.name
        state.state.devices.get(device.name)


async def load_and_schedule_polldevs(polldevs_conf: str):
    new_devices, deleted_devices = load_polldevs(polldevs_conf)
    schedule_new_devices(new_devices)
    deschedule_deleted_devices(deleted_devices)


def schedule_new_devices(new_devices: Sequence[str]):
    devices = sorted((state.polldevs[name] for name in new_devices), key=operator.attrgetter("priority"), reverse=True)
    if not devices:
        return

    _log.debug("Scheduling %s new devices", len(devices))

    scheduler = get_scheduler()

    # Spread poll jobs evenly across the entire default interval
    stagger_factor = (DEFAULT_INTERVAL_MINUTES * 60) / len(devices)
    for index, device in enumerate(devices):
        first_run_time = datetime.now() + timedelta(seconds=index * stagger_factor)

        scheduler.add_job(
            func=run_all_tasks,
            trigger="interval",
            minutes=device.interval,
            args=(device, state.state),
            next_run_time=first_run_time,
            name=device.name,
            id=device.name,
        )


def deschedule_deleted_devices(deleted_devices: Sequence[str]):
    """De-schedules recurring jobs for the deleted devices"""
    scheduler = get_scheduler()
    for name in deleted_devices:
        try:
            scheduler.remove_job(job_id=name)
        except JobLookupError:
            _log.debug("Job for device %s could not be found", name)



import logging
from asyncio import AbstractEventLoop
from typing import Optional

from zino.api.legacy import Zino1ServerProtocol
from zino.api.notify import Zino1NotificationProtocol
from zino.state import ZinoState
from zino.statemodels import Event

_logger = logging.getLogger(__name__)


class ZinoServer:
    """Represents the two asyncio servers that work in tandem to implement the Zino 1 legacy API:

    Port 8001 is the text-based command interface.
    Port 8002 is the text-based notification interface.
    """

    API_PORT = 8001
    NOTIFY_PORT = 8002

    def __init__(self, loop: AbstractEventLoop, state: ZinoState):
        self._loop = loop
        self.state: ZinoState = state
        self.active_clients: set[Zino1ServerProtocol] = set()
        self.notification_channels: dict[str, Zino1NotificationProtocol] = {}
        self.notify_server = self.api_server = None

    def serve(self, address: str = "0.0.0.0"):
        """Sets up the two asyncio servers to serve in tandem 'forever'"""
        api_coroutine = self._loop.create_server(
            lambda: Zino1ServerProtocol(server=self, state=self.state), address, self.API_PORT
        )
        self.api_server = self._loop.run_until_complete(api_coroutine)
        _logger.info("Serving API on %r", self.api_server.sockets[0].getsockname())

        notify_coroutine = self._loop.create_server(
            lambda: Zino1NotificationProtocol(server=self, state=self.state), address, self.NOTIFY_PORT
        )
        self.notify_server = self._loop.run_until_complete(notify_coroutine)
        _logger.info("Serving notifications on %r", self.notify_server.sockets[0].getsockname())

        self.state.events.add_event_observer(self.on_event_commit)

    def on_event_commit(self, new_event: Event, old_event: Optional[Event] = None) -> None:
        """Event observer to build notifications for notification channels"""
        Zino1NotificationProtocol.build_and_send_notifications(self, new_event, old_event)



import logging
from unittest.mock import Mock, patch

import pytest
from apscheduler.jobstores.base import JobLookupError

from zino import scheduler
from zino.state import ZinoState


class TestLoadPolldevs:
    @patch("zino.state.polldevs", dict())
    @patch("zino.state.state", ZinoState())
    def test_should_return_all_new_devices_on_first_run(self, polldevs_conf):
        new_devices, deleted_devices = scheduler.load_polldevs(polldevs_conf)
        assert len(new_devices) > 0
        assert not deleted_devices

    @patch("zino.state.polldevs", dict())
    @patch("zino.state.state", ZinoState())
    def test_should_return_deleted_devices_on_second_run(self, polldevs_conf, polldevs_conf_with_single_router):
        scheduler.load_polldevs(polldevs_conf)
        new_devices, deleted_devices = scheduler.load_polldevs(polldevs_conf_with_single_router)
        assert not new_devices
        assert len(deleted_devices) > 0

    @patch("zino.state.polldevs", dict())
    @patch("zino.state.state", ZinoState())
    def test_should_return_no_new_or_deleted_devices_on_invalid_configuration(self, invalid_polldevs_conf):
        new_devices, deleted_devices = scheduler.load_polldevs(invalid_polldevs_conf)
        assert not new_devices
        assert not deleted_devices

    @patch("zino.state.polldevs", dict())
    @patch("zino.state.state", ZinoState())
    def test_should_log_error_on_invalid_configuration(self, caplog, invalid_polldevs_conf):
        with caplog.at_level(logging.ERROR):
            scheduler.load_polldevs(invalid_polldevs_conf)
        assert "'lalala' is not a valid configuration line" in caplog.text


class TestScheduleNewDevices:
    @patch("zino.state.polldevs", dict())
    @patch("zino.state.state", ZinoState())
    def test_should_schedule_jobs_for_new_devices(self, polldevs_conf, mocked_scheduler):
        new_devices, _ = scheduler.load_polldevs(polldevs_conf)
        assert len(new_devices) > 0

        scheduler.schedule_new_devices(new_devices)
        assert mocked_scheduler.add_job.called

    @patch("zino.state.polldevs", dict())
    def test_should_do_nothing_when_device_list_is_empty(self, mocked_scheduler):
        scheduler.schedule_new_devices([])
        assert not mocked_scheduler.add_job.called


def test_deschedule_deleted_devices_should_deschedule_jobs(mocked_scheduler):
    scheduler.deschedule_deleted_devices(["test-gw"])
    assert mocked_scheduler.remove_job.called


def test_deschedule_deleted_devices_should_not_fail_on_not_found_job(mocked_scheduler_raising_error, caplog):
    with caplog.at_level(logging.DEBUG):
        scheduler.deschedule_deleted_devices(["test-gw"])
    assert mocked_scheduler_raising_error.remove_job.called
    assert "Job for device test-gw could not be found" in caplog.text


def test_scheduler_should_be_initialized_without_error():
    sched = scheduler.get_scheduler()
    assert sched


@pytest.fixture
def mocked_scheduler():
    with patch("zino.scheduler.get_scheduler") as get_scheduler:
        mock_scheduler = Mock()
        get_scheduler.return_value = mock_scheduler

        yield mock_scheduler


@pytest.fixture
def mocked_scheduler_raising_error():
    with patch("zino.scheduler.get_scheduler") as get_scheduler:
        mock_scheduler = Mock()
        mock_scheduler.remove_job.side_effect = Mock(side_effect=JobLookupError(job_id="fake id"))
        get_scheduler.return_value = mock_scheduler

        yield mock_scheduler



import ipaddress
import logging
from asyncio import Future
from datetime import timedelta
from unittest.mock import Mock

import pytest

from zino import flaps
from zino.config.models import PollDevice
from zino.flaps import (
    FLAP_CEILING,
    FLAP_INIT_VAL,
    FLAP_MIN,
    FLAP_THRESHOLD,
    FlappingState,
    FlappingStates,
    age_flapping_states,
    age_single_interface_flapping_state,
    stabilize_flapping_state,
)
from zino.state import ZinoState
from zino.statemodels import EventState, FlapState, Port, PortStateEvent
from zino.tasks.linkstatetask import LinkStateTask
from zino.time import now


class TestFlappingState:
    def test_update_should_increase_histval(self):
        flapping_state = FlappingState()
        initial = flapping_state.hist_val

        flapping_state.update()
        assert flapping_state.hist_val > initial

    def test_update_should_set_last_flap_timestamp(self):
        flapping_state = FlappingState()
        initial = now() - timedelta(seconds=10)
        flapping_state.last_flap = initial

        flapping_state.update()
        assert flapping_state.last_flap > initial

    def test_update_should_never_increase_hist_val_above_ceiling(self):
        flapping_state = FlappingState(hist_val=FLAP_CEILING)

        flapping_state.update()
        assert flapping_state.hist_val <= FLAP_CEILING

    def test_age_should_decrease_hist_val(self):
        flapping_state = FlappingState(hist_val=42)

        flapping_state.age()
        assert flapping_state.hist_val < 42

    def test_when_hist_val_is_below_threshold_is_below_threshold_should_return_true(self):
        flapping_state = FlappingState(hist_val=FLAP_THRESHOLD - 0.2)

        assert flapping_state.is_below_threshold()


class TestFlappingStates:
    def test_update_interface_flap_should_update_existing_flapping_state(self):
        flapping_states = FlappingStates()
        flapping_states.interfaces[1] = FlappingState(hist_val=42)

        flapping_states.update_interface_flap(1)
        assert flapping_states.interfaces[1].hist_val > 42

    def test_when_interface_has_no_flapping_stats_update_interface_flap_should_initialize_them(self):
        flapping_states = FlappingStates()

        flapping_states.update_interface_flap(1)
        assert 1 in flapping_states.interfaces

    def test_first_flap_should_initialize_flapping_state(self):
        flapping_states = FlappingStates()

        flapping_states.first_flap(1)
        assert flapping_states.interfaces[1].flaps == 1
        assert flapping_states.interfaces[1].hist_val == FLAP_INIT_VAL

    def test_unflap_should_remove_flapping_state(self):
        flapping_states = FlappingStates()
        flapping_states.interfaces[1] = FlappingState()

        flapping_states.unflap(1)
        assert 1 not in flapping_states.interfaces

    def test_when_hist_val_is_above_threshold_is_flapping_should_set_flag(self):
        flapping_states = FlappingStates()
        flapping_states.interfaces[1] = FlappingState(hist_val=FLAP_THRESHOLD + 0.2)

        flapping_states.is_flapping(1)
        assert flapping_states.interfaces[1].flapped_above_threshold

    def test_when_hist_val_is_above_threshold_is_flapping_should_return_true(self):
        flapping_states = FlappingStates()
        flapping_states.interfaces[1] = FlappingState(hist_val=FLAP_THRESHOLD + 0.2)

        assert flapping_states.is_flapping(1)

    def test_when_hist_val_is_below_minimal_threshold_is_flapping_should_return_false(self):
        flapping_states = FlappingStates()
        flapping_states.interfaces[1] = FlappingState(hist_val=FLAP_MIN - 0.2)

        assert not flapping_states.is_flapping(1)

    def test_when_flapping_stats_is_between_thresholds_is_flapping_should_return_false(self):
        flapping_states = FlappingStates()
        flapping_states.interfaces[1] = FlappingState(hist_val=int((FLAP_THRESHOLD + FLAP_MIN) / 2))

        assert not flapping_states.is_flapping(1)

    def test_when_flapping_stats_do_not_exist_is_flapping_should_return_false(self):
        flapping_states = FlappingStates()

        assert not flapping_states.is_flapping(1)

    def test_when_flapping_stats_exist_was_flapping_should_return_true(self):
        flapping_states = FlappingStates()
        flapping_states.interfaces[1] = FlappingState()

        assert flapping_states.was_flapping(1)

    def test_when_flapping_stats_do_not_exist_was_flapping_should_return_false(self):
        flapping_states = FlappingStates()

        assert not flapping_states.was_flapping(1)

    def test_when_flapping_stats_exist_get_flap_count_should_return_flap_count(self):
        flapping_states = FlappingStates()
        flapping_states.interfaces[1] = FlappingState(flaps=42)

        assert flapping_states.get_flap_count(1) == 42

    def test_when_flapping_stats_do_not_exist_get_flap_count_should_return_zero(self):
        flapping_states = FlappingStates()

        assert flapping_states.get_flap_count(1) == 0

    def test_when_flapping_stats_exist_get_flap_value_should_return_hist_val(self):
        flapping_states = FlappingStates()
        flapping_states.interfaces[1] = FlappingState(hist_val=42)

        assert flapping_states.get_flap_value(1) == 42

    def test_when_flapping_stats_do_not_exist_get_flap_value_should_return_zero(self):
        flapping_states = FlappingStates()

        assert flapping_states.get_flap_value(1) == 0


class TestAgeSingleInterfaceFlappingState:
    @pytest.mark.asyncio
    async def test_it_should_decrease_hist_val(self, state_with_flapstats, polldevs_dict):
        port: Port = next(iter(state_with_flapstats.devices.devices["localhost"].ports.values()))
        flapping_state = state_with_flapstats.flapping.interfaces[("localhost", port.ifindex)]
        initial = flapping_state.hist_val

        await age_single_interface_flapping_state(
            flapping_state, ("localhost", port.ifindex), state=state_with_flapstats, polldevs=polldevs_dict
        )
        assert flapping_state.hist_val < initial

    @pytest.mark.asyncio
    async def test_when_flap_is_below_threshold_it_should_remove_flapping_state(
        self, mocked_out_poll_single_interface, state_with_flapstats, polldevs_dict
    ):
        port: Port = next(iter(state_with_flapstats.devices.devices["localhost"].ports.values()))
        flapping_state = state_with_flapstats.flapping.interfaces[("localhost", port.ifindex)]
        flapping_state.hist_val = FLAP_THRESHOLD

        await age_single_interface_flapping_state(
            flapping_state, ("localhost", port.ifindex), state=state_with_flapstats, polldevs=polldevs_dict
        )

        assert not state_with_flapstats.flapping.interfaces


class TestStabilizeFlappingState:
    @pytest.mark.asyncio
    async def test_when_no_event_exists_it_should_create_an_event(
        self, mocked_out_poll_single_interface, state_with_flapstats, polldevs_dict
    ):
        port: Port = next(iter(state_with_flapstats.devices.devices["localhost"].ports.values()))
        flapping_state = state_with_flapstats.flapping.interfaces[("localhost", port.ifindex)]
        flapping_state.hist_val = FLAP_THRESHOLD

        assert len(state_with_flapstats.events.events) == 0

        await stabilize_flapping_state(
            flapping_state, ("localhost", port.ifindex), state=state_with_flapstats, polldevs=polldevs_dict
        )

        assert len(state_with_flapstats.events.events) > 0

    @pytest.mark.asyncio
    async def test_when_a_matching_event_exists_it_should_set_its_flapstate_to_stable(
        self, mocked_out_poll_single_interface, state_with_flapstats, polldevs_dict
    ):
        port: Port = next(iter(state_with_flapstats.devices.devices["localhost"].ports.values()))
        flapping_state = state_with_flapstats.flapping.interfaces[("localhost", port.ifindex)]
        flapping_state.hist_val = FLAP_THRESHOLD

        orig_event = state_with_flapstats.events.get_or_create_event("localhost", port.ifindex, PortStateEvent)
        orig_event.flapstate = FlapState.FLAPPING
        orig_event.flaps = 42
        orig_event.port = port.ifdescr
        orig_event.portstate = port.state
        orig_event.router = "localhost"
        orig_event.polladdr = "127.0.0.1"
        orig_event.priority = 500
        orig_event.ifindex = port.ifindex
        orig_event.descr = port.ifalias

        state_with_flapstats.events.commit(orig_event)

        await stabilize_flapping_state(
            flapping_state, ("localhost", port.ifindex), state=state_with_flapstats, polldevs=polldevs_dict
        )

        updated_event = state_with_flapstats.events.get_or_create_event("localhost", port.ifindex, PortStateEvent)
        assert updated_event
        assert updated_event.flapstate == FlapState.STABLE

    @pytest.mark.asyncio
    async def test_when_a_matching_closed_event_exists_it_should_set_its_flapstate_to_stable(
        self, mocked_out_poll_single_interface, state_with_flapstats, polldevs_dict
    ):
        port: Port = next(iter(state_with_flapstats.devices.devices["localhost"].ports.values()))
        flapping_state = state_with_flapstats.flapping.interfaces[("localhost", port.ifindex)]
        flapping_state.hist_val = FLAP_THRESHOLD

        orig_event = state_with_flapstats.events.get_or_create_event("localhost", port.ifindex, PortStateEvent)
        orig_event.flapstate = FlapState.FLAPPING
        orig_event.flaps = 42
        orig_event.port = port.ifdescr
        orig_event.portstate = port.state
        orig_event.router = "localhost"
        orig_event.polladdr = "127.0.0.1"
        orig_event.priority = 500
        orig_event.ifindex = port.ifindex
        orig_event.descr = port.ifalias

        state_with_flapstats.events.commit(orig_event)
        orig_event.set_state(EventState.CLOSED)
        state_with_flapstats.events.commit(orig_event)

        await stabilize_flapping_state(
            flapping_state, ("localhost", port.ifindex), state=state_with_flapstats, polldevs=polldevs_dict
        )

        closed_event = state_with_flapstats.events.get_closed_event("localhost", port.ifindex, PortStateEvent)
        assert closed_event
        assert closed_event.flapstate == FlapState.STABLE

    @pytest.mark.asyncio
    async def test_when_port_is_unkown_it_should_still_log_the_change(
        self,
        mocked_out_poll_single_interface,
        state_with_flapstats,
        polldevs_dict,
        caplog,
    ):
        port: Port = next(iter(state_with_flapstats.devices.devices["localhost"].ports.values()))
        flapping_state = state_with_flapstats.flapping.interfaces[("localhost", port.ifindex)]
        flapping_state.hist_val = FLAP_THRESHOLD
        fake_ifindex = port.ifindex + 666

        with caplog.at_level(logging.INFO):
            await stabilize_flapping_state(
                flapping_state, ("localhost", fake_ifindex), state=state_with_flapstats, polldevs=polldevs_dict
            )

        assert f"{fake_ifindex} stopped flapping" in caplog.text


@pytest.mark.asyncio
async def test_age_flapping_states_should_age_all_flapping_states(monkeypatch, event_loop):
    future = event_loop.create_future()
    future.set_result(None)
    mock_ager = Mock(return_value=future)
    monkeypatch.setattr(flaps, "age_single_interface_flapping_state", mock_ager)

    state = ZinoState()
    state.flapping.first_flap(("localhost", 1))
    state.flapping.first_flap(("localhost", 2))

    await age_flapping_states(state, {})
    assert mock_ager.call_count == 2


@pytest.fixture
def state_with_flapstats(state_with_localhost_with_port) -> ZinoState:
    initial = FLAP_THRESHOLD * 2
    port: Port = next(iter(state_with_localhost_with_port.devices.devices["localhost"].ports.values()))
    last_change = now() - timedelta(minutes=1)
    flapping_state = FlappingState(
        hist_val=initial,
        first_flap=last_change - timedelta(minutes=10),
        last_flap=last_change,
        last_age=last_change,
    )
    flapstates = FlappingStates(interfaces={("localhost", port.ifindex): flapping_state})
    state_with_localhost_with_port.flapping = flapstates
    return state_with_localhost_with_port


@pytest.fixture
def polldevs_dict(polldevs_conf_with_single_router) -> dict[str, PollDevice]:
    return {"localhost": PollDevice(name="localhost", address=ipaddress.IPv4Address("127.0.0.1"))}


@pytest.fixture
def mocked_out_poll_single_interface(monkeypatch):
    """Monkey patches LinkStateTask.poll_single_interface to do essentially nothing"""
    future = Future()
    future.set_result(None)
    monkeypatch.setattr(LinkStateTask, "poll_single_interface", future)
    yield monkeypatch



import pytest

from zino.statemodels import ReachabilityEvent, ReachabilityState
from zino.tasks.errors import DeviceUnreachableError


class TestReachableTask:
    @pytest.mark.asyncio
    async def test_run_should_not_create_event_if_device_is_reachable(self, reachable_task):
        task = reachable_task
        assert (await task.run()) is None
        event = task.state.events.get(task.device.name, None, ReachabilityEvent)
        assert not event

    @pytest.mark.asyncio
    async def test_run_should_create_event_if_device_is_unreachable(self, unreachable_task):
        task = unreachable_task
        with pytest.raises(DeviceUnreachableError):
            await task.run()
        event = task.state.events.get(task.device.name, None, ReachabilityEvent)
        assert event

    @pytest.mark.asyncio
    async def test_run_should_start_extra_job_if_device_is_unreachable(self, unreachable_task):
        with pytest.raises(DeviceUnreachableError):
            await unreachable_task.run()
        assert unreachable_task._extra_job_is_running()

    @pytest.mark.asyncio
    async def test_run_should_not_start_extra_job_if_device_is_reachable(self, reachable_task):
        task = reachable_task
        assert (await task.run()) is None
        assert not task._extra_job_is_running()

    @pytest.mark.asyncio
    async def test_run_should_update_event_to_reachable_when_device_is_reachable(self, reachable_task):
        task = reachable_task
        event = task.state.events.create_event(task.device.name, None, ReachabilityEvent)
        event.reachability = ReachabilityState.NORESPONSE
        task.state.events.commit(event)

        assert (await task.run()) is None
        updated_event = task.state.events[event.id]
        assert updated_event.reachability == ReachabilityState.REACHABLE

    @pytest.mark.asyncio
    async def test_run_should_update_event_to_noresponse_when_device_is_unreachable(self, unreachable_task):
        task = unreachable_task
        event = task.state.events.create_event(task.device.name, None, ReachabilityEvent)
        event.reachability = ReachabilityState.REACHABLE
        task.state.events.commit(event)

        with pytest.raises(DeviceUnreachableError):
            await task.run()
        updated_event = task.state.events[event.id]
        assert updated_event.reachability == ReachabilityState.NORESPONSE

    @pytest.mark.asyncio
    async def test_run_extra_job_should_update_event_to_reachable_when_device_is_reachable(self, reachable_task):
        task = reachable_task
        event = task.state.events.create_event(task.device.name, None, ReachabilityEvent)
        event.reachability = ReachabilityState.NORESPONSE
        task.state.events.commit(event)

        assert (await task._run_extra_job()) is None
        updated_event = task.state.events[event.id]
        assert updated_event.reachability == ReachabilityState.REACHABLE

    @pytest.mark.asyncio
    async def test_run_extra_job_should_not_update_event_when_device_is_unreachable(self, unreachable_task):
        task = unreachable_task
        event = task.state.events.create_event(task.device.name, None, ReachabilityEvent)
        event.reachability = ReachabilityState.NORESPONSE
        assert (await task._run_extra_job()) is None
        assert event.reachability == ReachabilityState.NORESPONSE



"""SNMP trap 'daemon' for Zino 2"""

import asyncio
import logging
from dataclasses import dataclass, field
from ipaddress import ip_address
from typing import Any, Dict, List, NamedTuple, Optional, Set

from pyasn1.type.base import SimpleAsn1Type
from pysnmp.carrier.asyncio.dgram import udp
from pysnmp.entity import config
from pysnmp.entity.rfc3413 import ntfrcv
from pysnmp.proto.rfc1902 import ObjectName
from pysnmp.smi import view
from pysnmp.smi.rfc1902 import ObjectIdentity, ObjectType

import zino.state
from zino.config.models import PollDevice
from zino.oid import OID
from zino.snmp import _mib_value_to_python, get_new_snmp_engine
from zino.statemodels import DeviceState, IPAddress

TrapType = tuple[str, str]  # A mib name and a corresponding trap symbolic name
_logger = logging.getLogger(__name__)


class TrapVarBind(NamedTuple):
    """Describes a single trap varbind as high-level as possible, but with low level details available as well"""

    oid: OID
    mib: str
    var: str
    instance: OID
    raw_value: Any
    value: Any


class TrapOriginator(NamedTuple):
    """Describes the originating SNMP agent of a trap message in Zino terms"""

    address: IPAddress
    port: int
    device: Optional[DeviceState] = None


@dataclass
class TrapMessage:
    """Describes an incoming trap message in the simplest possible terms needed for Zino usage"""

    agent: TrapOriginator
    mib: Optional[str] = None
    name: Optional[str] = None
    variables: List[TrapVarBind] = field(default_factory=list)

    def __str__(self):
        variables = [f"{v.mib}::{v.var}{v.instance or ''}={v.value or v.raw_value}" for v in self.variables]
        variables = ", ".join(variables)
        return f"<Trap from {self.agent.device.name}: {variables}>"

    def __contains__(self, label) -> bool:
        for var in self.variables:
            if var.var == label:
                return True
        return False

    def get_all(self, label: str) -> List[TrapVarBind]:
        """Returns all contained variables with the given label"""
        return [var for var in self.variables if var.var == label]


class TrapObserver:
    """Defines a valid protocol for SNMP trap observers.

    A trap observer that directly subclasses this protocol can expect to be automatically registered by Zino as an
    observer for any trap it declares in its `WANTED_TRAPS` attribute.
    """

    WANTED_TRAPS: Set[TrapType] = set()

    def __init__(
        self,
        state: zino.state.ZinoState,
        polldevs: Optional[Dict[str, PollDevice]] = None,
        loop: Optional[asyncio.AbstractEventLoop] = None,
    ):
        """Initializes a new trap observer with a reference to Zino's state.

        Optionally also receives a reference to the current polldevs configuration and the running event loop.
        """
        self.state = state
        self.polldevs: Dict[str, PollDevice] = polldevs or zino.state.polldevs
        self.loop = loop if loop else asyncio.get_event_loop()

    async def handle_trap(self, trap: TrapMessage) -> Optional[bool]:
        """A trap observer receives a trap message and an optional event loop reference.  The event loop reference
        may be useful if the observer needs to run async actions as part of its trap processing.  If the trap
        observer returns a true-ish value, the trap dispatcher will offer the same trap to more subscribers.  If a
        false-ish value is returned, the trap dispatcher will stop processing this trap.
        """
        ...


class TrapReceiver:
    """Zino Adapter for SNMP trap reception using PySNMP.

    A major difference to Zino 1 is that this receiver must explicitly be configured with SNMP community strings that
    will be accepted.  Zino 1 accepts traps with any community string, as long as their origin is any one of the
    devices configured in the pollfile.  However, PySNMP places heavy emphasis on being standards compliant,
    and will not even pass on traps to our callbacks unless they match the authorization config for the SNMP engine.
    """

    def __init__(
        self,
        address: str = "0.0.0.0",
        port: int = 162,
        loop: Optional[asyncio.AbstractEventLoop] = None,
        state: Optional[zino.state.ZinoState] = None,
        polldevs: Optional[Dict[str, PollDevice]] = None,
    ):
        self.transport: udp.UdpTransport = None
        self.address = address
        self.port = port
        self.loop = loop if loop else asyncio.get_event_loop()
        self.state = state or zino.state.ZinoState()
        self.polldevs = polldevs if polldevs is not None else {}
        self.snmp_engine = get_new_snmp_engine()
        self._communities = set()
        self._observers: dict[TrapType, List[TrapObserver]] = {}
        self._auto_subscribed_observers = set()

    def auto_subscribe_observers(self):
        """Automatically subscribes all loaded TrapObserver subclasses to this trap receiver"""
        for observer_class in TrapObserver.__subclasses__():
            if not observer_class.WANTED_TRAPS:
                continue
            if observer_class in self._auto_subscribed_observers:
                continue
            else:
                self._auto_subscribed_observers.add(observer_class)
            observer_instance = observer_class(state=self.state, polldevs=self.polldevs, loop=self.loop)
            self.observe(observer_instance, *observer_instance.WANTED_TRAPS)

    def observe(self, subscriber: TrapObserver, *trap_types: List[TrapType]):
        """Adds a trap subscriber to the receiver"""
        for trap_type in trap_types:
            observers = self._observers.setdefault(trap_type, [])
            observers.append(subscriber)

    def get_observers_for(self, trap_type: TrapType) -> List[TrapObserver]:
        """Returns a list of trap observers for a given trap type"""
        return self._observers.get(trap_type, [])

    async def open(self):
        """Opens the UDP transport socket and starts receiving traps"""
        self.transport = udp.UdpTransport(loop=self.loop).openServerMode((self.address, self.port))
        # This attribute needs to be awaited to ensure the socket is really opened,
        # unsure of why PySNMP doesn't do this, or how else it's supposed to be achieved
        await self.transport._lport
        _logger.info("Listening for incoming SNMP traps on %r", (self.address, self.port))
        config.addTransport(self.snmp_engine, udp.domainName + (1,), self.transport)

        ntfrcv.NotificationReceiver(self.snmp_engine, self.trap_received)
        self.snmp_engine.transportDispatcher.jobStarted(1)  # this job would never finish

    def close(self):
        """Closes the running SNMP engine and its associated ports"""
        self.snmp_engine.transportDispatcher.closeDispatcher()
        self.transport.closeTransport()

    def add_community(self, community: str):
        """Adds a new community string that will be accepted on incoming packets"""
        if community in self._communities:
            return
        self._communities.add(community)
        config.addV1System(self.snmp_engine, str(len(self._communities)), community)

    def trap_received(self, snmp_engine, state_reference, context_engine_id, context_name, var_binds, callback_context):
        """Callback function that receives all matched trap messages on PySNMP's incoming transport socket"""
        transport_domain, transport_address = snmp_engine.msgAndPduDsp.getTransportInfo(state_reference)
        sender_address, sender_port = transport_address
        sender_address = ip_address(sender_address)

        router = self._lookup_device(sender_address)
        if not router:
            _logger.info("ignored trap from %s (not a box we monitor?)", sender_address)
            return

        _logger.debug(
            'Trap from %s (%s) ContextEngineId "%s", ContextName "%s"',
            router.name,
            sender_address,
            context_engine_id.prettyPrint(),
            context_name.prettyPrint(),
        )
        origin = TrapOriginator(address=sender_address, port=sender_port, device=router)
        trap = TrapMessage(agent=origin)
        snmp_trap_oid: ObjectName = None
        for var, raw_value in var_binds:
            mib, label, instance, raw_value = self._resolve_varbind(var, raw_value)
            _logger.debug("(%r, %r, %s) = %s", mib, label, instance, raw_value.prettyPrint())
            try:
                value = _mib_value_to_python(raw_value)
            except Exception:  # noqa
                value = None
            trap.variables.append(TrapVarBind(OID(var), mib, label, instance, raw_value, value))
            if label == "snmpTrapOID":
                snmp_trap_oid = raw_value

        if not self._verify_trap(trap):
            return

        # TODO do some time calculations, but ask Håvard what the deal is with RestartTime vs. BootTime

        trap.mib, trap.name, _ = self._resolve_object_name(snmp_trap_oid)
        asyncio.ensure_future(self.dispatch_trap(trap))

    @staticmethod
    def _verify_trap(trap: TrapMessage) -> bool:
        device = trap.agent.device.name if trap.agent.device else "N/A"
        if "snmpTrapOID" not in trap:
            _logger.error("Trap from %s did not contain a snmpTrapOID value, ignoring", device)
            return False

        if "sysUpTime" not in trap:
            _logger.error("Trap from %s did not contain a sysUpTime value, ignoring", device)
            return False

        return True

    async def dispatch_trap(self, trap: TrapMessage):
        """Dispatches incoming trap messages according to internal subscriptions"""
        observers = self.get_observers_for((trap.mib, trap.name))
        if not observers:
            _logger.info("unknown trap: %s", trap)
            return

        for observer in observers:
            try:
                if not await observer.handle_trap(trap):
                    return
            except Exception:  # noqa
                _logger.exception("Unhandled exception in trap observer %r", observer)

    def _lookup_device(self, address: IPAddress) -> Optional[DeviceState]:
        """Looks up a device from Zino's running state from an IP address"""
        name = self.state.addresses.get(address)
        if name in self.state.devices:
            return self.state.devices[name]

    def _resolve_object_name(self, object_name: ObjectName) -> tuple[str, str, OID]:
        """Raises MibNotFoundError if oid in `object_name` can not be found"""
        engine = self.snmp_engine
        controller = engine.getUserContext("mibViewController")
        if not controller:
            controller = view.MibViewController(engine.getMibBuilder())
        mib, label, instance = controller.getNodeLocation(object_name)
        return mib, label, OID(instance) if instance else None

    def _resolve_varbind(self, name: ObjectName, value: SimpleAsn1Type) -> tuple[str, str, OID, SimpleAsn1Type]:
        """Resolves a varbind name and value to a MIB, label, instance, and value.  The value will be interpreted
        according to the resolved MIB object.

        Raises MibNotFoundError if the object's MIB cannot be resolved.
        """
        engine = self.snmp_engine
        controller = engine.getUserContext("mibViewController")
        if not controller:
            controller = view.MibViewController(engine.getMibBuilder())

        name, value = ObjectType(ObjectIdentity(name), value).resolveWithMib(controller)
        mib, label, instance = name.getMibSymbol()
        return mib, label, instance, value



import asyncio
from ipaddress import ip_address
from typing import Optional

import aiodns
from pyasn1.type.univ import OctetString

from zino.statemodels import IPAddress


def parse_ip(ip: str) -> IPAddress:
    """Parses IPv4 and IPv6 addresses in different formats"""
    try:
        return ip_address(ip)
    except ValueError:
        if ip.startswith("0x"):
            return _parse_hexa_string_ip(ip)
        if ":" in ip:
            return _parse_colon_separated_ip(ip)
        raise ValueError(f"Input {ip} could not be converted to IP address.")


def _parse_hexa_string_ip(ip: str) -> IPAddress:
    """Parses IP addresses formatted as hexastrings, e.g. 0x7f000001"""
    return ip_address(bytes(OctetString(hexValue=ip[2:])))


def _parse_colon_separated_ip(ip: str) -> IPAddress:
    """Parses IP addresses formatted with a colon symbol separating every octet, e.g. 7F:00:00:01"""
    return ip_address(bytes(OctetString(hexValue=ip.replace(":", ""))))


async def reverse_dns(ip: str) -> Optional[str]:
    """Returns hostname for given IP address or None if reverse DNS lookup fails"""
    loop = asyncio.get_event_loop()
    resolver = aiodns.DNSResolver(loop=loop)
    try:
        response = await resolver.gethostbyaddr(ip)
        return response.name
    except aiodns.error.DNSError:
        return None



import logging
from datetime import datetime, timedelta, timezone
from typing import get_args

from zino.events import EventIndex
from zino.state import ZinoState
from zino.stateconverter.linedata import LineData
from zino.stateconverter.utils import OldState, parse_ip, parse_log_and_history
from zino.statemodels import (
    AlarmEvent,
    AlarmType,
    BFDEvent,
    BFDSessState,
    BGPAdminStatus,
    BGPEvent,
    BGPOperState,
    EventState,
    InterfaceState,
    PortStateEvent,
    ReachabilityEvent,
    ReachabilityState,
    SubIndex,
)

_log = logging.getLogger(__name__)


event_name_to_type = {
    "bgp": BGPEvent,
    "bfd": BFDEvent,
    "reachability": ReachabilityEvent,
    "alarm": AlarmEvent,
    "portstate": PortStateEvent,
}


EventIndices = dict[str, EventIndex]


def set_event_state(
    old_state: OldState,
    new_state: ZinoState,
):
    event_indices: EventIndices = {}
    for linedata in old_state["::EventIdToIx"]:
        event_id, event_index = _get_event_index(linedata)
        event_indices[event_id] = event_index
    for linedata in old_state["::EventAttrs"]:
        try:
            _set_event_attrs(linedata, new_state, event_indices)
        except ValueError as e:
            _log.error(f"Error setting event attribute: {str(e)}")
    new_state.events._rebuild_indexes()


def _get_event_index(line: LineData) -> tuple[int, EventIndex]:
    """Parses a LineData representing an EventIdToIx line and returns a tuple
    of event_id, event_index
    """
    event_id = int(line.identifiers[0])

    index_components = line.value.split(",")
    if len(index_components) not in [2, 3]:
        raise ValueError(f"Invalid event index {line.value}")
    device, *rest = index_components
    # The last component should always be the event type
    event_type = rest.pop() if rest else None
    # If anything is left, it should be the subindex
    subindex = _parse_subindex(rest[0]) if rest else None

    event_index = EventIndex(device, subindex, event_name_to_type[event_type])
    return event_id, event_index


def _parse_subindex(subindex: str) -> SubIndex:
    """Parses the part of a EventIdToIx line that defines the ip/port value"""
    if subindex is None:
        return subindex
    if subindex in get_args(AlarmType):
        return subindex
    try:
        return parse_ip(subindex)
    except ValueError:
        pass
    try:
        return int(subindex)
    except ValueError:
        pass
    raise ValueError(f"Invalid SubIndex: {subindex}")


def _set_event_attrs(linedata: LineData, state: ZinoState, indices: EventIndices):
    event_field = linedata.identifiers[0]
    event_id = int(linedata.identifiers[1])
    event_index = indices[event_id]
    state.events.last_event_id = max(state.events.last_event_id, event_id)
    if event_id in state.events.events:
        event = state.events.events[event_id]
    else:
        event = state.events.create_event(*event_index)
        event.id = event_id
    if event_field == "priority":
        event.priority = int(linedata.value)
    elif event_field == "history":
        event.history = parse_log_and_history(linedata.value)
    elif event_field == "bgpOS":
        event.bgpos = BGPOperState(linedata.value)
    elif event_field == "bgpAS":
        event.bgpas = BGPAdminStatus(linedata.value)
    elif event_field == "lastevent":
        event.lastevent = linedata.value
    elif event_field == "log":
        event.log = parse_log_and_history(linedata.value)
    elif event_field == "polladdr":
        event.polladdr = parse_ip(linedata.value)
    elif event_field == "opened":
        event.opened = datetime.fromtimestamp(int(linedata.value), tz=timezone.utc)
    elif event_field == "peer-uptime":
        event.peer_uptime = int(linedata.value)
    elif event_field == "remote-AS":
        event.remote_as = int(linedata.value)
    elif event_field == "remote-addr":
        event.remote_addr = parse_ip(linedata.value)
    elif event_field == "router":
        event.router = linedata.value
    elif event_field == "state":
        event.state = EventState(linedata.value)
    elif event_field == "updated":
        event.updated = datetime.fromtimestamp(int(linedata.value), tz=timezone.utc)
    elif event_field == "ac-down":
        event.ac_down = timedelta(seconds=int(linedata.value))
    elif event_field == "descr":
        event.descr = linedata.value
    elif event_field == "flaps":
        _log.info("flaps is not a supported event field")
    elif event_field == "flapstate":
        _log.info("flapstate is not a supported event field")
    elif event_field == "ifindex":
        event.ifindex = int(linedata.value)
    elif event_field == "portstate":
        event.portstate = InterfaceState(linedata.value)
    elif event_field == "port":
        event.port = linedata.value
    elif event_field == "bfdAddr":
        if "unknown" in linedata.value:
            pass
        else:
            event.bfdaddr = parse_ip(linedata.value)
    elif event_field == "bfdDiscr":
        event.bfddiscr = int(linedata.value)
    elif event_field == "bfdIx":
        event.bfdix = int(linedata.value)
    elif event_field == "bfdState":
        event.bfdstate = BFDSessState(linedata.value)
    elif event_field == "lasttrans":
        event.lasttrans = datetime.fromtimestamp(int(linedata.value), tz=timezone.utc)
    elif event_field == "alarm-count":
        event.alarm_count = int(linedata.value)
    elif event_field == "alarm-type":
        event.alarm_type = linedata.value
    elif event_field == "Neigh-rDNS":
        event.neigh_rdns = linedata.value
    elif event_field == "reachability":
        event.reachability = ReachabilityState(linedata.value)
    elif event_field == "reason":
        event.reason = linedata.value
    elif event_field in ["id", "type"]:
        # These are set via other means
        pass
    else:
        raise ValueError(f"Unknown event attribute {event_field}")
    state.events.events[event.id] = event



from ipaddress import IPv4Address, IPv6Address
from unittest.mock import AsyncMock, MagicMock

import aiodns
import pytest

from zino.utils import parse_ip, reverse_dns


class TestParseIP:
    def test_should_parse_normal_ipv4_string_correctly(self):
        ip_string = "127.0.0.1"
        ip = parse_ip(ip_string)
        assert ip == IPv4Address(ip_string)

    def test_should_parse_normal_ipv6_string_correctly(self):
        ip_string = "13c7:db1c:4430:c826:6333:aed0:e605:3a3b"
        ip = parse_ip(ip_string)
        assert ip == IPv6Address(ip_string)

    def test_should_parse_hexastring_ipv4_correctly(self):
        ip = parse_ip("0x7f000001")
        assert ip == IPv4Address("127.0.0.1")

    def test_should_parse_hexastring_ipv6_correctly(self):
        ip = parse_ip("0x13c7db1c4430c8266333aed0e6053a3b")
        assert ip == IPv6Address("13c7:db1c:4430:c826:6333:aed0:e605:3a3b")

    def test_should_parse_colon_separated_ipv4_octets_correctly(self):
        ip = parse_ip("7f:00:00:01")
        assert ip == IPv4Address("127.0.0.1")

    def test_should_parse_colon_separated_ipv6_octets_correctly(self):
        ip = parse_ip("13:c7:db:1c:44:30:c8:26:63:33:ae:d0:e6:05:3a:3b")
        assert ip == IPv6Address("13c7:db1c:4430:c826:6333:aed0:e605:3a3b")

    def test_should_raise_valueerror_if_invalid_ip_format(self):
        with pytest.raises(ValueError):
            parse_ip("invalidformat")

    def test_should_raise_error_if_invalid_hexastring(self):
        with pytest.raises(ValueError):
            parse_ip("0xinvalidstring")

    def test_should_raise_error_if_invalid_colon_separated_string(self):
        with pytest.raises(ValueError):
            parse_ip(":thisis:just:randomstuff")


class TestReverseDNS:
    @pytest.mark.asyncio
    async def test_should_return_reverse_dns_for_valid_ip(self, mock_dnsresolver):
        valid_ip = "8.8.8.8"
        reverse_dns_value = "reverse.dns.example.com"

        # Create mock for return value of gethostbyaddr
        mock_gethostbyaddr_return_value = MagicMock()
        mock_gethostbyaddr_return_value.configure_mock(name=reverse_dns_value, aliases=[], addresses=[valid_ip])

        # Create gethostbyaddr async mock
        mock_gethostbyaddr = AsyncMock(return_value=mock_gethostbyaddr_return_value)

        # Mock DNSRsolver.gethostbyaddr function
        mock_dnsresolver.gethostbyaddr = mock_gethostbyaddr

        result = await reverse_dns(valid_ip)
        assert result == reverse_dns_value

    @pytest.mark.asyncio
    async def test_should_return_none_for_invalid_ip(self, mock_dnsresolver):
        invalid_ip = "0.0.0.0"

        # Mock DNSResolver.gethostbyaddr to raise DNSError
        mock_gethostbyaddr = AsyncMock(side_effect=aiodns.error.DNSError)
        mock_dnsresolver.gethostbyaddr = mock_gethostbyaddr

        result = await reverse_dns(invalid_ip)
        assert result is None


@pytest.fixture
def mock_dnsresolver(monkeypatch) -> AsyncMock:
    mock_dnsresolver = AsyncMock()
    monkeypatch.setattr("zino.utils.aiodns.DNSResolver", lambda loop: mock_dnsresolver)
    return mock_dnsresolver



"""This module implements custom logging of certain SNMP traps, but takes no other action"""

import logging
from typing import Optional

from zino.trapd import TrapMessage, TrapObserver

_logger = logging.getLogger(__name__)


class RestartTrapLogger(TrapObserver):
    WANTED_TRAPS = {
        ("SNMPv2-MIB", "coldStart"),
        ("SNMPv2-MIB", "warmStart"),
    }

    async def handle_trap(self, trap: TrapMessage) -> Optional[bool]:
        _logger.info("%s: %s", trap.agent.device.name, trap.name)
        return False  # stop trap processing here



import pytest
from pydantic import ValidationError

from zino.config import InvalidConfigurationError, read_configuration


class TestReadConfiguration:
    def test_returns_config_defined_in_file(self, zino_conf, polldevs_conf_with_no_routers):
        config = read_configuration(zino_conf)
        assert config
        assert config.polling.file == str(polldevs_conf_with_no_routers)

    def test_pollfile_argument_overrides_pollfile_defined_in_file(self, zino_conf, polldevs_conf_with_single_router):
        config = read_configuration(config_file_name=zino_conf, poll_file_name=str(polldevs_conf_with_single_router))
        assert config
        assert config.polling.file == str(polldevs_conf_with_single_router)

    def test_raises_error_on_file_not_found(self):
        with pytest.raises(OSError):
            read_configuration("non-existent-config.toml")

    def test_raises_error_on_invalid_toml_file(self, invalid_zino_conf):
        with pytest.raises(InvalidConfigurationError):
            read_configuration(invalid_zino_conf)

    def test_raises_error_on_invalid_config_values(self, invalid_values_zino_conf):
        with pytest.raises(ValidationError):
            read_configuration(invalid_values_zino_conf)

    def tests_raises_error_on_misspelled_key(self, extra_keys_zino_conf):
        with pytest.raises(ValidationError):
            read_configuration(extra_keys_zino_conf)

    def test_raises_error_on_pollfile_not_found(self, tmp_path):
        pollfile = "non-existent-pollfile.toml"
        name = tmp_path / pollfile
        with open(name, "w") as conf:
            conf.write(
                f"""
                [polling]
                file = "{pollfile}"
                """
            )
        with pytest.raises(ValidationError):
            read_configuration(name)
